#
# Copyright (c) Daniel Sheffield 2021
#
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
from collections import OrderedDict
import urwid
from decimal import Decimal
from urwid import numedit

import app
from app.db_utils import NON_IDENTIFIER_COLUMNS

class AutoCompleteEdit(urwid.Edit):
    def __init__(self, name, *args, apply_change_func=None, **kwargs):
        if isinstance(name, tuple):
            pallete, title = name
            self.name = title
            title = title.title()
            passthrough = (pallete, u'  ' if title.lower() == 'unit' else u'')
        else:
            self.name = name
            title = name.title()
            passthrough = u'  ' if name.lower() == 'unit' else u''
            
        super().__init__(passthrough, *args, **kwargs)
        self.apply = apply_change_func

    def keypress(self, size, key):
        if key == 'enter':
            self.apply(self.name)
            return
        elif key == 'delete':
            self.set_edit_text('')
            return
        
        return super().keypress(size, key)


class AutoCompleteFloatEdit(numedit.FloatEdit):
    def __init__(self, name, *args, apply_change_func=None, **kwargs):
        self.last_val = None
        self.op = '='
        self.pallete = None
        if isinstance(name, tuple):
            self.pallete, self.name = name
            title = self.name.title()
            passthrough = (self.pallete, f'{self.op} ')
        else:
            self.name = name
            title = name.title()
            passthrough = title
            
        super(AutoCompleteFloatEdit, self).__init__(passthrough, *args, **kwargs)
        self.apply = apply_change_func

    def update_caption(self):
        if self.pallete is not None:
            self.set_caption((self.pallete, f'{self.op} '))
        else:
            self.set_caption(f'{self.op} ')

    def set_op(self, op):
        self.op = op
        self.last_val = self.value()
        self.set_edit_text('')
        self.update_caption()
        
    def calc(self):
        x = self.last_val
        op = self.op
        if op in ('+', '-',):
            y = self.value() or Decimal(0.0)
            if op == '+':
                z = x + y
            else:
                z = x - y
        elif op in ('*', '/'):
            y = self.value() or Decimal(1.0)
            if op == '*':
                z = x * y
            else:
                z = x / y
        else:
            y = self.value() or Decimal(0.0)
            z = y
        
        self.op = '='
        self.update_caption()
        self.set_edit_text(f'{z:.2f}')

    def keypress(self, size, key):
        if isinstance(key, tuple):
            return
        ops = ('+', '-', '*', '/',)

        if key in ops:
            if self.op in ops:
                self.calc()
            self.set_op(key)
            return
        elif key == 'enter':
            if self.get_edit_text() == '' or self.value() == Decimal(0.0):
                return self.apply(self.name)
            self.calc()
            return
        elif key == '=':
            self.calc()
            return
        elif key == 'delete':
            self.set_edit_text('')
            self.op = '='
            self.update_caption()
            return

        return super().keypress(size, key)

class NoTabCheckBox(urwid.CheckBox):
    def keypress(self, size, key):
        if not isinstance(key, tuple) and key == 'tab':
            return 
        else:
            return super().keypress(size, key)

class FocusWidget(urwid.WidgetPlaceholder):

    def __init__(self, widget, initial_focus, skip_focus):
        super().__init__(widget)
        self._initial_focus = tuple([ i for i in initial_focus ])
        self._skip_focus = tuple([ i for i in skip_focus ])
    
    @property
    def skip_focus(self):
        return self._skip_focus
    
    @property
    def initial_focus(self):
        return list(self._initial_focus)
    
    @property
    def container(self):
        return self.original_widget.original_widget.original_widget
    
    def _set_focus_path(self, path):
        try:
            self.container.set_focus_path(path)
            return
        except IndexError:
            pass
        
        if path[-1] == 0 and len(path) > 1:
            self._set_focus_path(path[:-1])
            return
    
        raise IndexError

    def iter_focus_paths(self):
        self._set_focus_path(self.initial_focus)
        while True:
            path = self.container.get_focus_path()
            yield path
            self.advance_focus()
            path = self.container.get_focus_path()
            if path == self.initial_focus:
                break

    def advance_focus(self, reverse=False):

        path = self.container.get_focus_path()
                
        if reverse:
            paths = [ i for i in self.iter_focus_paths() ]
            zipped_paths = zip(paths, [
                *paths[1:], paths[0]
            ])
            prev_path = map(lambda x: x[0], filter(
                lambda x: x[1] == path,
                zipped_paths
            ))
            p = next(prev_path)
            self._set_focus_path(p)
            return
        
        _iter = [ i for i in enumerate(path) ][::-1]

        for idx, part in _iter:
            p = [ i for i in path ]
            if reverse:
                p[idx] -= 1
            else:
                p[idx] += 1
            
            try:
                self._set_focus_path(p)
                if p in self.skip_focus:
                    self.advance_focus(reverse=reverse)
                return
            except IndexError:
                path[idx] = 0
            
        self.container.set_focus_path(self.initial_focus)

class SuggestionPopup(urwid.Overlay):
    
    def __init__(self, under, name, options, apply_cb, esc_cb):
        self.esc_cb = esc_cb
        self.under = under
        body = [urwid.Text(name.title()), urwid.Divider()]
        for c in options:
            button = urwid.Button(c)
            urwid.connect_signal(button, 'click', apply_cb, c)
            body.append(urwid.AttrMap(button, None, focus_map='reversed'))
        walker = urwid.SimpleFocusListWalker(body, wrap_around=False)
        listbox = urwid.ListBox(walker)
        pad = urwid.Padding(listbox, left=2, right=2)
        pad = urwid.AttrMap(pad, 'banner')
        super().__init__(pad, under,
            align='center', width=('relative', 60),
            valign='middle', height=('relative', 60),
            min_width=20, min_height=9)

    def keypress(self, size, key):
        if key == 'esc':
            self.esc_cb()
            return

        if key == 'tab':
            return

        return super().keypress(size, key)

class NewProduct(urwid.Overlay):

    def __init__(self, query_manager, under, name, data, autocomplete_cb, change_cb, apply_cb, esc_cb):
        self.esc_cb = esc_cb
        self.under = under
        self._data = data
        self.query_manager = query_manager
        self.name = name
        
        title = urwid.Text('Enter Product Info', align='center')
        self.fields = OrderedDict()
        for f in ('product', 'category', 'group'):
            w = AutoCompleteEdit(
                    ('bg', f),
                    apply_change_func=lambda name: autocomplete_cb(name)
                )
            self.fields[f] = w
            w.set_edit_text(data[f])
            urwid.connect_signal(w, 'change', change_cb(f))

        ok = urwid.Button('Done', on_press=lambda w: apply_cb(**self.data))

        body = urwid.AttrMap(urwid.ListBox(urwid.SimpleListWalker([
                urwid.Pile([
                    urwid.AttrMap(title, 'banner'),
                    urwid.Divider(),
                    *[
                        urwid.AttrMap(
                            urwid.LineBox(urwid.AttrMap(v,'streak'), title=k.title(), title_align='left'), 'banner'
                        ) for k,v in self.fields.items()
                    ],
                    urwid.AttrMap(ok, 'banner'),
                ], focus_item=2)
            ])
        ), 'banner')
        super().__init__(urwid.AttrMap(body, 'bg'), under,
            align='center', width=('relative', 40),
            valign='middle', height=('relative', 40),
            min_width=20, min_height=12)

    @property
    def data(self):
        return dict([(k,v) for k,v in self._data.items() if k in ('product', 'category', 'group')])
    
    def _apply_choice(self, name, value):
        self._data.update({
            name: value
        })
        for k,v in self.data.items():
            if k == name or v:
                continue
            options = self.query_manager.unique_suggestions(k, exclude=[self.name, *NON_IDENTIFIER_COLUMNS], **self.data)
            if len(options) == 1 and k != 'ts':
                self._data.update({
                    k: list(options)[0]
                })
    
    def apply_choice(self, name):
        return  lambda w,x: self._apply_choice(name, x)

    def update(self):
        for k, v in self.data.items():
            self.fields[k].set_edit_text(v)
        return self

    def keypress(self, size, key):
        if key == 'esc':
            self.esc_cb()
            return

        if key == 'tab':
            return

        return super().keypress(size, key)
